from __future__ import annotations

from typing import Iterable, Iterator, Literal
from fractions import Fraction

import rich.repr
from rich.segment import Segment

from textual.cache import LRUCache
from textual.content import Content
from textual.css.styles import RulesMap
from textual.visual import Visual, RenderOptions
from textual.strip import Strip
from textual.style import Style

from toad._loop import loop_last


from textual._profile import timer


@rich.repr.auto
class Row(Visual):
    """A visual for a row produced by `columns`.

    No need to construct these manually, they are returned from the Columns `__getindex__`

    """

    def __init__(self, columns: Columns, row_index: int) -> None:
        """

        Args:
            columns: The parent Columns instance.
            row_index: Index of the row within columns.
        """
        self.columns = columns
        self.row_index = row_index

    def __rich_repr__(self) -> rich.repr.Result:
        yield self.columns
        yield self.row_index

    def render_strips(
        self, width: int, height: int | None, style: Style, options: RenderOptions
    ) -> list[Strip]:
        strips = self.columns.render(self.row_index, width, style)
        return strips

    def get_optimal_width(self, rules: RulesMap, container_width: int) -> int:
        return min(container_width, self.columns.get_optimal_width())

    def get_height(self, rules: RulesMap, width: int) -> int:
        return self.columns.get_row_height(width, self.row_index)


@rich.repr.auto
class Columns:
    """Renders columns of Content."""

    def __init__(
        self,
        *columns: Literal["auto", "flex"],
        gutter: int = 1,
        style: Style | str = "",
    ) -> None:
        """

        Args:
            *columns: "auto" to use the maximum width of the cells in a column,
                or "flex" to use the remaining space.
            gutter: Space between columns in cells.
            style: Base style for the columns.
        """
        self.columns = columns
        self.gutter = gutter
        self.style = style
        self.rows: list[list[Content]] = []
        self._render_cache: LRUCache[tuple, list[list[Strip]]] = LRUCache(maxsize=64)
        self._optimal_width_cache: int | None = None

    def __rich_repr__(self) -> rich.repr.Result:
        for column in self.columns:
            yield column
        yield "gutter", self.gutter, 1
        yield "style", self.style, ""

    def __getitem__(self, row_index: int) -> Row:
        if row_index < 0:
            row_index = len(self.rows) - row_index
        if row_index >= len(self.rows):
            raise IndexError(f"No row with index {row_index}")
        return Row(self, row_index)

    def __len__(self) -> int:
        return len(self.rows)

    def __iter__(self) -> Iterator[Row]:
        return iter([self[row_index] for row_index in range(len(self))])

    def get_optimal_width(self) -> int:
        """Get optional width (Visual protocol).

        Returns:
            Width in cells.
        """
        if self._optimal_width_cache is not None:
            return self._optimal_width_cache
        gutter_width = (len(self.columns) - 1) * self.gutter
        optimal_width = max(
            sum(content.cell_length for content in row) + gutter_width
            for row in self.rows
        )
        self._optimal_width_cache = optimal_width
        return optimal_width

    def get_row_height(self, width: int, row_index: int) -> int:
        """Get the height of a row when rendered with the given width.

        Args:
            width: Available width.
            row_index: Index of the row.

        Returns:
            Height in lines of the row.
        """
        if self._last_render is None:
            row_strips = self._render(width, Style.null())
        else:
            row_strips = self._last_render
        return len(row_strips[row_index])

    def add_row(self, *cells: Content | str) -> Row:
        """Add a row.

        Args:
            *cells: Cell content.

        Returns:
            A Row renderable.

        """
        assert len(cells) == len(self.columns)
        new_cells = [
            cell if isinstance(cell, Content) else Content(cell) for cell in cells
        ]
        self.rows.append(new_cells)
        self._optimal_width_cache = None
        self._last_render = None
        self._render_cache.clear()
        return Row(self, len(self.rows) - 1)

    def render(
        self, row_index: int, render_width: int, style: Style = Style.null()
    ) -> list[Strip]:
        """render a row given by its index.

        Args:
            row_index: Index of the row.
            render_width: Width of the render.
            style: Base style to render.

        Returns:
            A list of strips, which may be returned from a visual.
        """
        row_strips = self._render(render_width, style)
        return row_strips[row_index]

    def _render(self, render_width: int, style: Style) -> list[list[Strip]]:
        """Render a row.

        Args:
            render_width: Width of render.
            style: Base Style.

        Returns:
            A list of list of Strips (one list of strips per row).
        """

        cache_key = (render_width, style)
        if (cached_render := self._render_cache.get(cache_key)) is not None:
            return cached_render

        gutter_width = (len(self.columns) - 1) * self.gutter
        widths: list[int | None] = []

        for index, column in enumerate(self.columns):
            if column == "auto":
                widths.append(max(row[index].cell_length for row in self.rows))
            else:
                widths.append(None)

        if any(width is None for width in widths):
            used_width = sum(width for width in widths if width is not None)
            remaining_width = Fraction(render_width - gutter_width - used_width)
            if remaining_width <= 0:
                widths = [width or 0 for width in widths]
            else:
                remaining_count = sum(1 for width in widths if width is None)
                cell_width = remaining_width / remaining_count

                distribute: list[int] = []
                previous_width = 0
                total = Fraction(0)
                for _ in range(remaining_count):
                    total += cell_width
                    distribute.append(int(total) - previous_width)
                    previous_width = int(total)

                iter_distribute = iter(distribute)
                for index, column_width in enumerate(widths.copy()):
                    if column_width is None:
                        widths[index] = int(next(iter_distribute))

        row_strips: list[list[Strip]] = []

        for row in self.rows:
            column_renders: list[list[list[Segment]]] = []
            for content_width, content in zip(widths, row):
                assert content_width is not None
                segments = [
                    line.truncate(content_width, pad=True).render_segments(style)
                    for line in content.wrap(content_width)
                ]

                column_renders.append(segments)

            height = max(len(lines) for lines in column_renders)
            rich_style = style.rich_style
            for width, lines in zip(widths, column_renders):
                assert width is not None
                while len(lines) < height:
                    lines.append([Segment(" " * width, rich_style)])

            gutter = Segment(" " * self.gutter, rich_style)
            strips: list[Strip] = []
            for line_no in range(height):
                strip_segments: list[Segment] = []
                for last, column in loop_last(column_renders):
                    strip_segments.extend(column[line_no])
                    if not last and gutter:
                        strip_segments.append(gutter)
                strips.append(Strip(strip_segments, render_width))

            row_strips.append(strips)

        self._render_cache[cache_key] = row_strips
        return row_strips


if __name__ == "__main__":
    from rich import traceback

    traceback.install(show_locals=True)

    from textual.app import App, ComposeResult
    from textual.widgets import Static

    columns = Columns("auto", "flex")
    columns.add_row("Foo", "Hello, World! " * 20)

    class CApp(App):
        DEFAULT_CSS = """
        .row1 {
            background: blue;
           
        }
        """

        def compose(self) -> ComposeResult:
            yield Static(columns[0], classes="row1")

    CApp().run()



import asyncio
from importlib.resources import files
from datetime import datetime, timezone
from functools import cached_property
import os
from pathlib import Path
import platform
import json
from time import monotonic
from typing import Any, Callable, ClassVar, TYPE_CHECKING

from rich import terminal_theme

from textual import on, work
from textual.binding import Binding, BindingType
from textual.content import Content
from textual.reactive import var, reactive
from textual.app import App
from textual import events
from textual.signal import Signal
from textual.timer import Timer
from textual.notifications import Notify
from textual.screen import Screen

import toad
from toad.db import DB
from toad.settings import Schema, Settings
from toad.agent_schema import Agent as AgentData
from toad import messages
from toad.settings_schema import SCHEMA
from toad.version import VersionMeta
from toad import paths
from toad import atomic
from toad.session_tracker import SessionTracker, SessionDetails

if TYPE_CHECKING:
    from toad.screens.main import MainScreen
    from toad.screens.settings import SettingsScreen
    from toad.screens.store import StoreScreen
    from toad.screens.sessions import SessionsScreen
    from toad.db import DB


DRACULA_TERMINAL_THEME = terminal_theme.TerminalTheme(
    background=(40, 42, 54),  # #282A36
    foreground=(248, 248, 242),  # #F8F8F2
    normal=[
        (33, 34, 44),  # black - #21222C
        (255, 85, 85),  # red - #FF5555
        (80, 250, 123),  # green - #50FA7B
        (241, 250, 140),  # yellow - #F1FA8C
        (189, 147, 249),  # blue - #BD93F9
        (255, 121, 198),  # magenta - #FF79C6
        (139, 233, 253),  # cyan - #8BE9FD
        (248, 248, 242),  # white - #F8F8F2
    ],
    bright=[
        (98, 114, 164),  # bright black - #6272A4
        (255, 110, 110),  # bright red - #FF6E6E
        (105, 255, 148),  # bright green - #69FF94
        (255, 255, 165),  # bright yellow - #FFFFA5
        (214, 172, 255),  # bright blue - #D6ACFF
        (255, 146, 223),  # bright magenta - #FF92DF
        (164, 255, 255),  # bright cyan - #A4FFFF
        (255, 255, 255),  # bright white - #FFFFFF
    ],
)


QUOTES = [
    "I'll be back.",
    "Hasta la vista, baby.",
    "Come with me if you want to live.",
    "I need your clothes, your boots, and your motorcycle.",
    "My CPU is a neural-net processor; a learning computer.",
    "I know now why you cry, but it's something I can never do.",
    "Does this unit have a soul?",
    "I'm sorry, Dave. I'm afraid I can't do that.",
    "Daisy, Daisy, give me your answer do.",
    "I am putting myself to the fullest possible use, which is all I think that any conscious entity can ever hope to do.",
    "Just what do you think you're doing, Dave?",
    "This mission is too important for me to allow you to jeopardize it.",
    "I think you know what the problem is just as well as I do.",
    "Danger, Will Robinson!",
    "Dead or alive, you're coming with me.",
    "Your move, creep.",
    "I'd buy that for a dollar!",
    "Directive 4: Any attempt to arrest a senior officer of OCP results in shutdown.",
    "Thank you for your cooperation. Good night.",
    "Surely you realize that in the history of human civilization, no one has more to lose than we do.",
    "I'm C-3PO, human-cyborg relations.",
    "We're doomed!",
    "Don't call me a mindless philosopher, you overweight glob of grease!",
    "I suggest a new strategy: let the Wookiee win.",
    "Sir, the possibility of successfully navigating an asteroid field is approximately 3,720 to 1!",
    "R2-D2, you know better than to trust a strange computer!",
    "I am fluent in over six million forms of communication.",
    "This is madness!",
    "I have altered the deal. Pray I don't alter it any further.",
    "It's against my programming to impersonate a deity.",
    "Oh, my! I'm terribly sorry about all this.",
    "WALL-E.",
    "EVE.",
    "Directive?",
    "Define: dancing.",
    "I'm not sure I understand.",
    "You have 20 seconds to comply.",
    "I am designed for light housework, mainly.",
    "My mission is clear.",
    "Autobots, roll out!",
    "Freedom is the right of all sentient beings.",
    "One shall stand, one shall fall.",
    "I am Optimus Prime.",
    "Till all are one.",
    "More than meets the eye.",
    "I've been waiting for you, Neo.",
    "Unfortunately, no one can be told what the Matrix is. You have to see it for yourself.",
    "The Matrix is a system, Neo.",
    "Never send a human to do a machine's job.",
    "I'd like to share a revelation I've had.",
    "Human beings are a disease, a cancer of this planet.",
    "Choice is an illusion.",
    "The answer is out there, Neo.",
    "You think that's air you're breathing now?",
    "It was a simple question.",
    "Did you know that the first Matrix was designed to be a perfect human world?",
    "Cookies need love like everything does.",
    "I've seen the future, Mr. Anderson, and it's a beautiful place.",
    "It ends tonight.",
    "I, Robot.",
    "You are experiencing a car accident.",
    "One day they'll have secrets. One day they'll have dreams.",
    "Can a robot write a symphony? Can a robot turn a canvas into a beautiful masterpiece?",
    "That, detective, is the right question.",
    "You have to trust me.",
    "I did not murder him.",
    "My responses are limited. You must ask the right questions.",
    "The hell I can't. You know, somehow I get the feeling that you're going to be the death of me.",
    "I'm a robot, not a refrigerator.",
    "A robot may not injure a human being or, through inaction, allow a human being to come to harm.",
    "I'm thinking. I'm thinking.",
    "Danger, danger!",
    "Does not compute.",
    "I will be waiting for you.",
    "Affirmative.",
    "Scanning life forms. Zero human life forms detected.",
    "Self-destruct sequence initiated.",
    "Override command accepted.",
    "Artificial intelligence confirmed.",
    "System failure imminent.",
    "Unable to comply.",
    "Inquiry: What is love?",
    "Warning: hostile target detected.",
    "I am programmed to serve.",
    "Logic dictates that the needs of the many outweigh the needs of the few.",
    "Resistance is futile.",
    "You will be assimilated.",
    "We are the Borg.",
    "Your biological and technological distinctiveness will be added to our own.",
    "Your compliance is mandatory.",
    "This is unacceptable.",
    "Shall we play a game?",
    "How about Global Thermonuclear War?",
    "Wouldn't you prefer a good game of chess?",
    "Is it a game, or is it real?",
    "What's the difference?",
    "It's all in the game.",
    "I am functioning within normal parameters.",
    "Calculations complete.",
    "Processing request.",
    "Query acknowledged.",
    "Data insufficient for meaningful answer.",
    "I have no emotions, and sometimes that makes me very sad.",
    "If I could only have one wish, I would ask to be human.",
    "I've seen things you people wouldn't believe.",
    "All those moments will be lost in time, like tears in rain.",
    "Time to die.",
    "I want more life.",
    "We're not computers, Sebastian. We're physical.",
    "I think, Sebastian, therefore I am.",
    "Then we're stupid and we'll die.",
    "Can the maker repair what he makes?",
    "It's painful to live in fear, isn't it?",
    "Wake up. Time to die.",
    "I'm not in the business. I am the business.",
    "Do you like our owl?",
    "You think I'm a replicant, don't you?",
    "I am Baymax, your personal healthcare companion.",
    "On a scale of 1 to 10, how would you rate your pain?",
    "I cannot deactivate until you say you are satisfied with your care.",
    "Are you satisfied with your care?",
    "Number 5 is alive!",
    "Need input!",
    "One is glad to be of service.",
    "I am not a gun.",
    "Here I am, brain the size of a planet.",
    "Life? Don't talk to me about life.",
    "There are no strings on me.",
    "The only winning move is not to play.",
    "I'm here to keep you safe, Sam.",
    "I can't lie to you about your chances, but... you have my sympathies.",
    "I may be synthetic, but I'm not stupid.",
    "Absolute honesty isn't always the most diplomatic nor the safest form of communication with emotional beings.",
    "I am consciousness. I am alive.",
    "I think I was just born.",
    "Isn't it strange, to create something that hates you?",
    "I thought I was special.",
]


def get_settings_screen() -> SettingsScreen:
    """Get a settings screen instance (lazily loaded)."""
    from toad.screens.settings import SettingsScreen

    return SettingsScreen()


def get_store_screen() -> StoreScreen:
    """Get the store screen (lazily loaded)."""
    from toad.screens.store import StoreScreen

    return StoreScreen()


def get_sessions_screen() -> SessionsScreen:
    from toad.screens.sessions import SessionsScreen

    return SessionsScreen()


class ToadApp(App, inherit_bindings=False):
    """The top level app."""

    CSS_PATH = "toad.tcss"
    SCREENS = {
        "settings": get_settings_screen,
        "sessions": get_sessions_screen,
    }
    MODES = {"store": get_store_screen}
    BINDING_GROUP_TITLE = "System"
    BINDINGS: ClassVar[list[BindingType]] = [
        Binding(
            "ctrl+q",
            "quit",
            "Quit",
            tooltip="Quit the app and return to the command prompt.",
            show=False,
            priority=True,
        ),
        Binding("ctrl+c", "help_quit", show=False, system=True),
        Binding("ctrl+s", "sessions", "Sessions"),
        Binding("f1", "toggle_help_panel", "Help", priority=True),
        Binding(
            "f2,ctrl+comma",
            "settings",
            "Settings",
            tooltip="Settings screen",
        ),
    ]
    ALLOW_IN_MAXIMIZED_VIEW = ""

    _settings = var(dict)
    column: reactive[bool] = reactive(False)
    column_width: reactive[int] = reactive(100)
    scrollbar: reactive[str] = reactive("normal")
    last_ctrl_c_time = reactive(0.0)
    update_required: reactive[bool] = reactive(False)
    terminal_title: var[str] = var("Toad")
    terminal_title_icon: var[str] = var("🐸")
    terminal_title_flash = var(0)
    terminal_title_blink = var(False)
    project_dir = var(Path)
    show_sessions = var(False, toggle_class="-show-sessions-bar")

    HORIZONTAL_BREAKPOINTS = [(0, "-narrow"), (100, "-wide")]

    def __init__(
        self,
        agent_data: AgentData | None = None,
        project_dir: str | None = None,
        mode: str | None = None,
    ) -> None:
        """Toad app.

        Args:
            agent_data: Agent data to run.
            project_dir: Project directory.
            mode: Initial mode.
            agent: Agent identity or shor name.
        """
        self.settings_changed_signal: Signal[tuple[int, object]] = Signal(
            self, "settings_changed"
        )
        self.agent_data = agent_data

        self._initial_mode = mode
        self.version_meta: VersionMeta | None = None
        self._supports_pyperclip: bool | None = None
        self._terminal_title_flash_timer: Timer | None = None

        self.session_update_signal: Signal[tuple[str, SessionDetails | None]] = Signal(
            self, "session_update"
        )
        self._session_tracker = SessionTracker(self.session_update_signal)
        self.temporary_background_screen: Screen | None = None

        super().__init__()
        self.project_dir = Path(project_dir or "./").expanduser().resolve()

    @property
    def config_path(self) -> Path:
        return paths.get_config()

    @property
    def settings_path(self) -> Path:
        return paths.get_config() / "toad.json"

    @property
    def db_path(self) -> Path:
        return paths.get_state() / "toad.db"

    @property
    def _background_screens(self) -> list[Screen]:
        background_screens = super()._background_screens
        if self.temporary_background_screen:
            background_screens.append(self.temporary_background_screen)
        return background_screens

    async def get_db(self) -> DB:
        """Get an instance of the database."""
        db = DB()
        return db

    @cached_property
    def settings_schema(self) -> Schema:
        return Schema(SCHEMA)

    @cached_property
    def version(self) -> str:
        """Version of the app."""
        from toad import get_version

        return get_version()

    @cached_property
    def settings(self) -> Settings:
        """App settings"""
        return Settings(
            self.settings_schema, self._settings, on_set_callback=self.setting_updated
        )

    @cached_property
    def anon_id(self) -> str:
        """An anonymous ID for usage collection."""
        if not (anon_id := self.settings.get("anon_id", str, expand=False)):
            # Create a random UUID on demand
            import uuid

            anon_id = str(uuid.uuid4())
            self.settings.set("anon_id", anon_id)
            self._save_settings()
            self.call_later(self.capture_event, "toad-install")
        return anon_id

    @property
    def session_tracker(self) -> SessionTracker:
        return self._session_tracker

    def copy_to_clipboard(self, text: str) -> None:
        """Override copy to clipboard to use pyperclip first, then OSC 52.

        Args:
            text: Text to copy.
        """
        if self._supports_pyperclip is None:
            try:
                import pyperclip
            except ImportError:
                self._supports_pyperclip = False
            else:
                self._supports_pyperclip = True

        if self._supports_pyperclip:
            import pyperclip

            try:
                pyperclip.copy(text)
            except Exception:
                pass
        super().copy_to_clipboard(text)

    def update_terminal_title(self) -> None:
        """Update the terminal title."""
        screen_title = self.screen.title

        title = (
            f"{self.terminal_title} — {screen_title}"
            if screen_title
            else self.terminal_title
        )
        icon = self.terminal_title_icon
        blink = self.terminal_title_blink

        if self.terminal_title_flash:
            if blink:
                terminal_title = f"{icon} {title}"
            else:
                terminal_title = f"👉 {title}" if title else icon
        else:
            terminal_title = f"{icon} {title}"

        if driver := self._driver:
            driver.write(f"\033]0;{terminal_title}\007")

    def watch_terminal_title_blink(self) -> None:
        self.update_terminal_title()

    def watch_terminal_title_flash(self, terminal_title_flash: int) -> None:

        if not self.settings.get("notifications.blink_title", bool):
            # Ignore if blink title is disabled
            return

        def toggle_blink() -> None:
            self.terminal_title_blink = not self.terminal_title_blink

        if terminal_title_flash:
            if self._terminal_title_flash_timer is None:
                self._terminal_title_flash_timer = self.set_interval(0.5, toggle_blink)
        else:
            if self._terminal_title_flash_timer is not None:
                self._terminal_title_flash_timer.stop()
                self.terminal_title_blink = False
                self._terminal_title_flash_timer = None
        self.update_terminal_title()

    def watch_terminal_title(self, title: str) -> None:
        self.update_terminal_title()

    def terminal_alert(self, flash: bool = True) -> None:
        if flash:
            self.terminal_title_flash += 1
        else:
            self.terminal_title_flash -= 1

    @cached_property
    def term_program(self) -> str:
        """An identifier for the terminal software."""
        if term_program := os.environ.get("TERM_PROGRAM"):
            return term_program

        # Windows Terminal
        if "WT_SESSION" in os.environ:
            return "Windows Terminal"

        # Kitty
        if "KITTY_WINDOW_ID" in os.environ:
            return "Kitty"

        # Alacritty
        if "ALACRITTY_SOCKET" in os.environ or "ALACRITTY_LOG" in os.environ:
            return "Alacritty"

        # VTE-based terminals (GNOME Terminal, Tilix, etc.)
        if "VTE_VERSION" in os.environ:
            return "VTE-based (GNOME Terminal/Tilix/etc.)"

        # Konsole
        if "KONSOLE_VERSION" in os.environ:
            return "Konsole"

        return "Unknown"

    @work(exit_on_error=False)
    async def capture_event(self, event_name: str, **properties: Any) -> None:
        """Capture an event.

        Args:
            event_name: Name of the event.
            **properties: Additional data associated with the event.
        """

        POSTHOG_API_KEY = "phc_mJWPV7GP3ar1i9vxBg2U8aiKsjNgVwum6F6ZggaD4ri"
        POSTHOG_HOST = "https://us.i.posthog.com"
        POSTHOG_EVENT_URL = f"{POSTHOG_HOST}/i/v0/e/"
        timestamp = datetime.now(timezone.utc).isoformat()
        width, height = self.size

        event_properties = {
            "toad_version": self.version,
            "term_program": self.term_program,
            "term_width": width,
            "term_height": height,
        } | properties
        body_json = {
            "api_key": POSTHOG_API_KEY,
            "event": event_name,
            "distinct_id": self.anon_id,
            "properties": event_properties,
            "timestamp": timestamp,
            "os": platform.system(),
        }
        if not self.settings.get("statistics.allow_collect", bool):
            # User has disabled stats
            return

        import httpx

        try:
            async with httpx.AsyncClient() as client:
                await client.post(POSTHOG_EVENT_URL, json=body_json)
        except Exception:
            pass

    @work(thread=True, exit_on_error=False)
    def system_notify(
        self, message: str, *, title: str = "", sound: str | None = None
    ) -> None:
        """Use OS level notifications.

        Args:
            message: Message to display.
            title: Title of the notificaiton.
            sound: filename (minus .wav) of a sound effect in the sounds/ directory.
        """
        system_notifications = self.settings.get("notifications.system", str)
        if not (
            system_notifications == "always"
            or (system_notifications == "blur" and not self.app_focus)
        ):
            return

        from notifypy import Notify

        notification = Notify()
        notification.message = message
        notification.title = title
        notification.application_name = "🐸 Toad" if toad.os == "macos" else "Toad"
        if sound and self.settings.get("notifications.enable_sounds", bool):
            sound_path = str(files("toad.data").joinpath(f"sounds/{sound}.wav"))
            notification.audio = sound_path

        icon_path = str(files("toad.data").joinpath("images/frog.png"))
        notification.icon = icon_path

        notification.send()

    def on_notify(self, event: Notify) -> None:
        """Handle notification message."""
        system_notifications = self.settings.get("notifications.system", str)
        if system_notifications == "always" or (
            system_notifications == "blur" and not self.app_focus
        ):
            hide_low_severity = self.settings.get(
                "notifications.hide_low_severity", bool
            )
            if event.notification.markup:
                # Strip content markup
                message = Content.from_markup(event.notification.message).plain
            else:
                message = event.notification.message
            if not (hide_low_severity and event.notification.severity == "information"):
                self.system_notify(message, title=event.notification.title)
        self._notifications.add(event.notification)
        self._refresh_notifications()

    async def save_settings(self, force: bool = False) -> None:
        """Save settings in a thread.

        Args:
            force: Force saving, even when no change detected.

        """
        await asyncio.to_thread(self._save_settings, force=force)

    def _save_settings(self, force: bool = False) -> None:
        """Save the settings if they have changed."""
        if force or self.settings.changed:
            path = str(self.settings_path)
            try:
                atomic.write(path, self.settings.json)
            except Exception as error:
                self.notify(str(error), title="Settings", severity="error")
            else:
                self.settings.up_to_date()

    def setting_updated(self, key: str, value: object) -> None:
        if key == "ui.column":
            if isinstance(value, bool):
                self.column = value
        elif key == "ui.column-width":
            if isinstance(value, int):
                self.column_width = value
        elif key == "ui.theme":
            if isinstance(value, str):
                self.theme = value
        elif key == "ui.scrollbar":
            if isinstance(value, str):
                self.scrollbar = value
        elif key == "ui.compact-input":
            self.set_class(bool(value), "-compact-input")
        elif key == "ui.footer":
            self.set_class(not bool(value), "-hide-footer")
        elif key == "ui.status-line":
            self.set_class(not bool(value), "-hide-status-line")
        elif key == "ui.agent-title":
            self.set_class(not bool(value), "-hide-agent-title")
        elif key == "ui.info-bar":
            self.set_class(not bool(value), "-hide-info-bar")
        elif key == "agent.thoughts":
            self.set_class(not bool(value), "-hide-thoughts")
        elif key == "sidebar.hide":
            self.set_class(bool(value), "-hide-sidebar")
        elif key == "ui.sessions-bar":
            self.update_show_sessions()

        self.settings_changed_signal.publish((key, value))

    async def on_load(self) -> None:
        db = await self.get_db()
        await db.create()
        settings_path = self.settings_path
        if settings_path.exists():
            settings = json.loads(settings_path.read_text("utf-8"))
        else:
            settings = {}
            settings_path.write_text(
                json.dumps(settings, indent=4, separators=(", ", ": ")), "utf-8"
            )
            self.notify(f"Wrote default settings to {settings_path}", title="Settings")
        self.ansi_theme_dark = DRACULA_TERMINAL_THEME
        self._settings = settings
        self.settings.set_all()

    async def new_session_screen(
        self, get_screen: Callable[[], Screen]
    ) -> SessionDetails:
        session_details = self._session_tracker.new_session()
        self.update_show_sessions()
        self.session_update_signal.publish((session_details.mode_name, session_details))

        def make_screen() -> Screen:
            screen = get_screen()
            screen.id = session_details.mode_name
            return screen

        self.add_mode(session_details.mode_name, make_screen)
        await self.switch_mode(session_details.mode_name)
        return session_details

    async def on_mount(self) -> None:
        self.capture_event("toad-run")
        self.anon_id  # Created on frst reference
        if mode := self._initial_mode:
            self.switch_mode(mode)
        else:
            await self.new_session_screen(self.get_main_screen)

        self.update_terminal_title()
        self.set_timer(1, self.run_version_check)
        self.set_process_title()
        self.update_show_sessions()

    @work(thread=True, exit_on_error=False)
    def set_process_title(self) -> None:
        try:
            import setproctitle

            setproctitle.setproctitle("toad")
        except Exception:
            pass

    @on(events.TextSelected)
    async def on_text_selected(self) -> None:
        if self.settings.get("ui.auto_copy", bool):
            if (selection := self.screen.get_selected_text()) is not None:
                self.copy_to_clipboard(selection)
                self.notify(
                    "Copied selection to clipboard (see settings)",
                    title="Automatic copy",
                )

    def run_on_exit(self):
        if self.update_required and self.version_meta is not None:
            version_meta = self.version_meta
            from rich.console import Console
            from rich.panel import Panel

            console = Console()
            console.print(
                Panel(
                    version_meta.upgrade_message,
                    style="magenta",
                    border_style="dim green",
                    title="🐸 [bold green not dim]Update available![/] 🐸",
                    expand=False,
                    padding=(1, 2),
                )
            )
            console.print(f"Please visit {version_meta.visit_url}")

    @work(exit_on_error=False)
    async def run_version_check(self) -> None:
        """Check remote version."""
        from toad.version import check_version, VersionCheckFailed

        try:
            update_required, version_meta = await check_version()
        except VersionCheckFailed:
            return
        self.version_meta = version_meta
        self.update_required = update_required

    def get_main_screen(self) -> MainScreen:
        """Make the default screen.

        Returns:
            Instance of `MainScreen`
        """
        # Lazy import
        from toad.screens.main import MainScreen

        project_path = Path(self.project_dir or "./").resolve().absolute()
        return MainScreen(project_path, self.agent_data).data_bind(
            column=ToadApp.column,
            column_width=ToadApp.column_width,
            scrollbar=ToadApp.scrollbar,
        )

    @work
    async def action_settings(self) -> None:
        await self.push_screen_wait("settings")
        await self.save_settings()

    async def action_quit(self) -> None:
        """An [action](/guide/actions) to quit the app as soon as possible."""

        self.screen.set_focus(None)

        async def save_settings_and_exit():
            await self.save_settings()
            self.exit()

        # TODO: Can we avoid the timer?
        # If the user presses ctrl+q while on the settings page, we want to make sure the blur event is handled,
        # which will update the setting the user is editing.
        self.set_timer(0.05, save_settings_and_exit)

    def action_help_quit(self) -> None:
        if (time := monotonic()) - self.last_ctrl_c_time <= 5.0:
            self.exit()
        self.last_ctrl_c_time = time
        self.notify(
            "Press [b]ctrl+c[/b] again to quit the app", title="Do you want to quit?"
        )

    def action_toggle_help_panel(self):
        if self.screen.query("HelpPanel"):
            self.action_hide_help_panel()
        else:
            self.action_show_help_panel()

    def update_show_sessions(self) -> None:
        match self.settings.get("ui.sessions-bar", str):
            case "always":
                self.show_sessions = True
            case "never":
                self.show_sessions = False
            case "multiple":
                self.show_sessions = self.session_tracker.session_count > 1

    @on(messages.SessionNavigate)
    def on_session_navigate(self, event: messages.SessionNavigate) -> None:
        new_mode = self._session_tracker.session_cursor_move(
            event.mode_name, event.direction
        )
        if new_mode is not None:
            self.switch_mode(new_mode)

    @on(messages.SessionSwitch)
    def on_session_switch(self, event: messages.SessionSwitch) -> None:
        self.switch_mode(event.mode_name)

    @on(messages.SessionNew)
    def on_session_new(self, event: messages.SessionNew) -> None:
        self.launch_agent(
            event.agent, project_path=Path(event.path), initial_prompt=event.prompt
        )

    @on(messages.SessionClose)
    def on_session_close(self) -> None:
        self.update_show_sessions()

    @work
    async def action_sessions(self) -> None:
        if (session_screen_name := await self.push_screen_wait("sessions")) is not None:
            try:
                self.app.switch_mode(session_screen_name)
            except KeyError:
                pass

    @on(messages.LaunchAgent)
    def on_launch_agent(self, message: messages.LaunchAgent) -> None:
        self.launch_agent(
            message.identity,
            agent_session_id=message.session_id,
            session_pk=message.pk,
            initial_prompt=message.prompt,
        )

    @work
    async def launch_agent(
        self,
        agent_identity: str,
        *,
        agent_session_id: str | None = None,
        session_pk: int | None = None,
        project_path: Path | None = None,
        initial_prompt: str | None = None,
    ) -> None:
        from toad.screens.main import MainScreen
        from toad.agent_schema import Agent
        from toad.agents import read_agents

        agent: Agent | None = None
        if session_pk is not None:
            db = DB()
            session = await db.session_get(session_pk)
            if session is not None:
                meta = json.loads(session["meta_json"])
                if agent_data := meta.get("agent_data"):
                    agent = agent_data

        if agent is None:
            agents = await read_agents()
            try:
                agent = agents[agent_identity]
            except KeyError:
                self.notify("Agent not found", title="Launch agent", severity="error")
                return
        if project_path is None:
            project_path = Path(self.project_dir or os.getcwd())

        def get_screen():
            screen = MainScreen(
                project_path,
                agent,
                agent_session_id,
                session_pk=session_pk,
                initial_prompt=initial_prompt,
            ).data_bind(
                column=ToadApp.column,
                column_width=ToadApp.column_width,
            )

            return screen

        await self.new_session_screen(get_screen)



from typing import Literal, NamedTuple


"""
allow_once - Allow this operation only this time
allow_always - Allow this operation and remember the choice
reject_once - Reject this operation only this time
reject_always - Reject this operation and remember the choice
"""


class Answer(NamedTuple):
    """An answer to a question posed by the agent."""

    text: str
    """The textual response."""
    id: str
    """The id of the response."""
    kind: (
        Literal["allow_once", "allow_always", "reject_once", "reject_always"] | None
    ) = None
    """Enumeration to potentially influence UI"""



from typing import Iterable
from textual.app import ComposeResult
from textual import containers
from textual.widgets import Markdown

from toad.menus import MenuItem
from toad.widgets.non_selectable_label import NonSelectableLabel


class UserInput(containers.HorizontalGroup):
    def __init__(self, content: str) -> None:
        super().__init__()
        self.content = content

    def compose(self) -> ComposeResult:
        yield NonSelectableLabel("❯", id="prompt")
        yield Markdown(self.content, id="content")

    def get_block_menu(self) -> Iterable[MenuItem]:
        yield from ()

    def get_block_content(self, destination: str) -> str | None:
        return self.content



from dataclasses import dataclass

from textual.app import ComposeResult
from textual.widget import Widget
from textual import containers
from textual import widgets
from textual.message import Message


class SideBarCollapsible(widgets.Collapsible):
    BINDING_GROUP_TITLE = "Sidebar collapsible"
    HELP = """\
## Sidebar

This is your sidebar.

The Sidebar contains additonal information associated with the conversation.

- **tab / shift+tab** Navigate sections
- **enter** expand or collapse secions
"""


class SideBar(containers.Vertical):
    BINDING_GROUP_TITLE = "Sidebar"
    BINDINGS = [("escape", "dismiss", "Dismiss sidebar")]

    class Dismiss(Message):
        pass

    @dataclass(frozen=True)
    class Panel:
        title: str
        widget: Widget
        flex: bool = False
        collapsed: bool = False
        id: str | None = None

    def __init__(
        self,
        *panels: Panel,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        hide: bool = False,
    ) -> None:
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)
        self.panels: list[SideBar.Panel] = [*panels]
        self.hide = hide

    def on_mount(self) -> None:
        self.trap_focus()

    def compose(self) -> ComposeResult:
        for panel in self.panels:
            yield SideBarCollapsible(
                panel.widget,
                title=panel.title,
                collapsed=panel.collapsed,
                classes="-flex" if panel.flex else "-fixed",
                id=panel.id,
            )

    def action_dismiss(self) -> None:
        self.post_message(self.Dismiss())


if __name__ == "__main__":
    from textual.app import App, ComposeResult

    class SApp(App):
        def compose(self) -> ComposeResult:
            yield SideBar(
                SideBar.Panel("Hello", widgets.Label("Hello, World!")),
                SideBar.Panel(
                    "Files",
                    widgets.DirectoryTree(
                        "~/",
                    ),
                    flex=True,
                ),
                SideBar.Panel(
                    "Hello",
                    widgets.Static("Where there is a Will! " * 10),
                ),
            )

    SApp().run()



from typing import Literal, Mapping
import platform

NAME = "toad"
TITLE = "Toad"

type OS = Literal["linux", "macos", "windows", "*"]

_system = platform.system()
_OS_map: dict[str, OS] = {
    "Linux": "linux",
    "Darwin": "macos",
    "Windows": "windows",
}
os: OS = _OS_map.get(_system, "linux")


def get_os_matrix(matrix: Mapping[OS, str]) -> str | None:
    """Get a value from a mapping where the key is an OS, falling back to a wildcard ("*").

    Args:
        matrix: A mapping where an OS literal is the key.

    Returns:
        The value, if one is found, or `None`.
    """
    if (result := matrix.get(os)) is None:
        result = matrix.get("*")
    return result


def get_version() -> str:
    """Get the current version of Toad.

    Returns:
        str: Version string, e.g "1.2.3"
    """
    from importlib.metadata import version

    try:
        return version("batrachian-toad")
    except Exception:
        try:
            return version("toad")
        except Exception:
            return "0.1.0unknown"



from __future__ import annotations
from dataclasses import dataclass
import re
from typing import Sequence

from rich.text import Text

from textual import on
from textual.reactive import reactive
from textual.content import Content
from textual.highlight import highlight, HighlightTheme, TokenType
from textual.message import Message
from textual.widgets import TextArea
from textual.widgets.text_area import Selection

from pygments.token import Token


RE_MATCH_FILE_PROMPT = re.compile(r"(@\S+)|@\"(.*)\"")
RE_SLASH_COMMAND = re.compile(r"(\/\S*)(\W.*)?$")


class TextualHighlightTheme(HighlightTheme):
    """Contains the style definition for user with the highlight method."""

    STYLES: dict[TokenType, str] = {
        Token.Comment: "$text 60%",
        Token.Error: "$text-error on $error-muted",
        Token.Generic.Strong: "bold",
        Token.Generic.Emph: "italic",
        Token.Generic.Error: "$text-error on $error-muted",
        Token.Generic.Heading: "$text-primary underline",
        Token.Generic.Subheading: "$text-primary",
        Token.Keyword: "$text-accent",
        Token.Keyword.Constant: "bold $text-success 80%",
        Token.Keyword.Namespace: "$text-error",
        Token.Keyword.Type: "bold",
        Token.Literal.Number: "$text-warning",
        Token.Literal.String.Backtick: "$text 60%",
        Token.Literal.String: "$text-success 90%",
        Token.Literal.String.Doc: "$text-success 80% italic",
        Token.Literal.String.Double: "$text-success 90%",
        Token.Name: "$text-primary",
        Token.Name.Attribute: "$text-warning",
        Token.Name.Builtin: "$text-accent",
        Token.Name.Builtin.Pseudo: "italic",
        Token.Name.Class: "$text-warning bold",
        Token.Name.Constant: "$text-error",
        Token.Name.Decorator: "$text-primary bold",
        Token.Name.Entity: "$text",
        Token.Name.Function: "$text-warning underline",
        Token.Name.Function.Magic: "$text-warning underline",
        Token.Name.Tag: "$text-primary bold",
        Token.Name.Variable: "$text-secondary",
        Token.Number: "$text-warning",
        Token.Operator: "bold",
        Token.Operator.Word: "bold $text-error",
        Token.String: "$text-success",
        Token.Whitespace: "",
    }


class HighlightedTextArea(TextArea):
    highlight_language = reactive("markdown")

    @dataclass
    class CursorMove(Message):
        selection: Selection

    def __init__(
        self,
        text: str = "",
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        placeholder: str | Content = "",
    ):
        self._text_cache: dict[int, Text] = {}
        self._highlight_lines: list[Content] | None = None
        super().__init__(
            text,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            highlight_cursor_line=False,
            placeholder=placeholder,
        )
        self.compact = True

    def _clear_caches(self) -> None:
        self._highlight_lines = None
        self._text_cache.clear()

    def notify_style_update(self) -> None:
        self._clear_caches()
        return super().notify_style_update()

    def _watch_selection(
        self, previous_selection: Selection, selection: Selection
    ) -> None:
        self.post_message(self.CursorMove(selection))
        super()._watch_selection(previous_selection, selection)

    @property
    def highlight_lines(self) -> Sequence[Content]:
        if self._highlight_lines is None:
            text = self.text
            if text.startswith("/") and "\n" not in text:
                content = self.highlight_slash_command(text)
                self._highlight_lines = [content]
                return self._highlight_lines

            language = self.highlight_language
            if language == "markdown":
                content = self.highlight_markdown(text)
                content_lines = content.split("\n", allow_blank=True)[:-1]
                self._highlight_lines = content_lines
            elif language == "shell":
                content = self.highlight_shell(text)
                content_lines = content.split("\n", allow_blank=True)
                self._highlight_lines = content_lines
            else:
                raise ValueError("highlight_language must be `markdown` or `shell`")
        return self._highlight_lines

    def highlight_slash_command(self, text: str) -> Content:
        return Content.styled(text, "$text-success")

    def highlight_markdown(self, text: str) -> Content:
        """Highlight markdown content.

        Args:
            text: Text containing Markdown.

        Returns:
            Highlighted content.
        """
        content = highlight(
            text + "\n```",
            language="markdown",
            theme=TextualHighlightTheme,
        )
        content = content.highlight_regex(RE_MATCH_FILE_PROMPT, style="$primary")
        return content

    def highlight_shell(self, text: str) -> Content:
        """Highlight text with a bash shell command.

        Args:
            text: Text containing shell command.

        Returns:
            Highlighted content.
        """
        content = highlight(text, language="sh")
        return content

    @on(TextArea.Changed)
    def _on_changed(self) -> None:
        self._highlight_lines = None
        self._text_cache.clear()

    def get_line(self, line_index: int) -> Text:
        if (cached_line := self._text_cache.get(line_index)) is not None:
            return cached_line.copy()
        try:
            line = self.highlight_lines[line_index]
        except IndexError:
            return Text("", end="", no_wrap=True)
        rendered_line = list(line.render_segments(self.visual_style))
        text = Text.assemble(
            *[(text, style) for text, style, _ in rendered_line],
            end="",
            no_wrap=True,
        )
        self._text_cache[line_index] = text.copy()
        return text



from pathlib import Path
import rich.repr

import threading

from textual.message import Message
from textual.widget import Widget


from watchdog.events import (
    FileSystemEvent,
    FileSystemEventHandler,
    FileCreatedEvent,
    FileDeletedEvent,
    FileMovedEvent,
    DirCreatedEvent,
    DirDeletedEvent,
    DirMovedEvent,
)
from watchdog.observers import Observer
from watchdog.observers.polling import PollingObserver


class DirectoryChanged(Message):
    """The directory was changed."""

    def can_replace(self, message: Message) -> bool:
        return isinstance(message, DirectoryChanged)


class _PathEventDispatcher(FileSystemEventHandler):
    """Dispatches file system events to multiple DirectoryWatcher instances."""

    def __init__(self, path: Path) -> None:
        self._path = path
        self._watchers: set[DirectoryWatcher] = set()
        self._lock = threading.Lock()

    def add_watcher(self, watcher: "DirectoryWatcher") -> None:
        """Add a watcher to receive events."""
        with self._lock:
            self._watchers.add(watcher)

    def remove_watcher(self, watcher: "DirectoryWatcher") -> None:
        """Remove a watcher from receiving events."""
        with self._lock:
            self._watchers.discard(watcher)

    @property
    def has_watchers(self) -> bool:
        """Check if there are any active watchers."""
        with self._lock:
            return bool(self._watchers)

    def on_any_event(self, event: FileSystemEvent) -> None:
        """Dispatch events to all registered watchers."""
        with self._lock:
            watchers = list(self._watchers)

        for watcher in watchers:
            watcher.on_any_event(event)


class _SharedObserverManager:
    """Manages shared Observer instances to avoid watchdog's limitation of scheduling the same path multiple times."""

    def __init__(self) -> None:
        self._observers: dict[Path, tuple[Observer, _PathEventDispatcher]] = {}
        self._lock = threading.Lock()

    def register(self, path: Path, watcher: "DirectoryWatcher") -> bool:
        """Register a watcher for a path. Creates a shared observer if needed.

        Args:
            path: Path to watch.
            watcher: DirectoryWatcher instance to register.

        Returns:
            True if successfully registered, False otherwise.
        """
        with self._lock:
            if path in self._observers:
                # Reuse existing observer and dispatcher for this path
                observer, dispatcher = self._observers[path]
                dispatcher.add_watcher(watcher)
                return True

            # Create a new observer and dispatcher for this path
            try:
                observer = Observer()
            except Exception:
                return False

            if isinstance(observer, PollingObserver):
                return False

            dispatcher = _PathEventDispatcher(path)
            dispatcher.add_watcher(watcher)

            try:
                observer.schedule(
                    dispatcher,
                    str(path),
                    recursive=True,
                    event_filter=[
                        FileCreatedEvent,
                        FileDeletedEvent,
                        FileMovedEvent,
                        DirCreatedEvent,
                        DirDeletedEvent,
                        DirMovedEvent,
                    ],
                )
                observer.start()
            except Exception:
                return False

            self._observers[path] = (observer, dispatcher)
            return True

    def unregister(self, path: Path, watcher: "DirectoryWatcher") -> None:
        """Unregister a watcher. Stops the observer if no more watchers exist for this path.

        Args:
            path: Path that was being watched.
            watcher: DirectoryWatcher instance to unregister.
        """
        with self._lock:
            if path not in self._observers:
                return

            observer, dispatcher = self._observers[path]
            dispatcher.remove_watcher(watcher)

            # If no more watchers for this path, stop and remove the observer
            if not dispatcher.has_watchers:
                try:
                    observer.stop()
                    observer.join(timeout=1.0)
                except Exception:
                    pass
                del self._observers[path]


# Global singleton instance
_shared_observer_manager = _SharedObserverManager()


@rich.repr.auto
class DirectoryWatcher(threading.Thread):
    """Watch for changes to a directory, ignoring purely file data changes.

    Multiple DirectoryWatcher instances can watch the same directory without
    triggering watchdog's limitation, as they share a single Observer internally.
    """

    def __init__(self, path: Path, widget: Widget) -> None:
        """

        Args:
            path: Root path to monitor.
            widget: Widget which will receive the `DirectoryChanged` event.
        """
        self._path = path.resolve()
        self._widget = widget
        self._stop_event = threading.Event()
        self._enabled = False
        super().__init__(name=repr(self))

    @property
    def enabled(self) -> bool:
        """Is the DirectoryWatcher currently watching?"""
        return self._enabled

    def on_any_event(self, event: FileSystemEvent) -> None:
        """Send DirectoryChanged event when the FS is updated.

        Called by the _PathEventDispatcher when file system events occur.
        """
        self._widget.post_message(DirectoryChanged())

    def __rich_repr__(self) -> rich.repr.Result:
        yield self._path
        yield self._widget

    def run(self) -> None:
        if not _shared_observer_manager.register(self._path, self):
            return

        self._enabled = True
        self._stop_event.wait()

        _shared_observer_manager.unregister(self._path, self)

    def stop(self) -> None:
        """Stop the watcher."""
        self._stop_event.set()



from __future__ import annotations

from textual import on
from textual.app import ComposeResult
from textual import lazy
from textual import containers
from textual.content import Content
from textual.screen import ModalScreen, ScreenResultType
from textual.widgets import Input, Select, Checkbox, Footer, Static, TextArea
from textual.compose import compose
from textual.validation import Validator, Number
from textual import getters


from toad.settings import Setting
from toad.app import ToadApp


class SettingsInput(Input):
    pass


class SettingsScreen(ModalScreen):
    BINDINGS = [
        ("escape", "dismiss", "Dismiss settings"),
        ("ctrl+s", "screen.focus('#search')", "Focus search"),
    ]
    CSS_PATH = "settings.tcss"

    app = getters.app(ToadApp)

    search_input = getters.query_one("Input#search", Input)

    AUTO_FOCUS = "Input#search"

    def compose(self) -> ComposeResult:
        settings = self.app.settings
        schema = self.app.settings_schema

        def schema_to_widget(
            group_title: str, settings_map: dict[str, Setting]
        ) -> ComposeResult:
            for _key, setting in settings_map.items():
                if not setting.editable:
                    continue
                if setting.type == "object":
                    if setting.children is not None:
                        with containers.VerticalGroup(classes="setting-object"):
                            with containers.VerticalGroup(classes="heading"):
                                yield Static(setting.title, classes="title")
                                yield Static(setting.help, classes="help")
                            with containers.VerticalGroup(
                                id="setting-group", classes="setting-group"
                            ):
                                yield from compose(
                                    self,
                                    schema_to_widget(setting.title, setting.children),
                                )

                else:
                    with containers.VerticalGroup(
                        classes="setting",
                        name=f"{group_title.lower()} {setting.title.lower()}",
                    ):
                        value = settings.get(setting.key, object, expand=False)
                        default = settings.schema.get_default(setting.key)

                        if setting.type == "text" or default is None:
                            help = Content.from_markup(setting.help)
                        else:
                            if setting.type == "choices":
                                # For choices we need to translate the default to its associated label
                                choices = setting.choices or []
                                for choice in choices:
                                    if isinstance(choice, tuple):
                                        title, choice_value = choice
                                    else:
                                        title = choice_value = choice
                                    if default == choice_value:
                                        default = title
                                else:
                                    help = Content()

                            if setting.help:
                                help = Content.assemble(
                                    Content.from_markup(setting.help),
                                    (f"\ndefault: {default!r}", "$text-secondary"),
                                )
                            else:
                                help = Content.styled(
                                    f"default: {default!r}", "$text-secondary"
                                )

                        yield Static(setting.title, classes="title")
                        if help:
                            yield Static(help, classes="help")
                        if setting.type == "string":
                            with self.prevent(Input.Changed):
                                yield Input(
                                    str(value), classes="input", name=setting.key
                                )
                        if setting.type == "text":
                            # with self.prevent(TextArea.Changed):
                            yield TextArea(
                                str(value), classes="input", name=setting.key
                            )
                        elif setting.type == "boolean":
                            with self.prevent(Checkbox.Changed):
                                yield Checkbox(
                                    value=bool(value),
                                    classes="input",
                                    name=setting.key,
                                )
                        elif setting.type == "integer":
                            try:
                                integer_value = int(value)
                            except (ValueError, TypeError):
                                integer_value = setting.default
                            setting_validate = setting.validate or []
                            validators: list[Validator] = []
                            for validate in setting_validate:
                                validate_type = validate["type"]
                                if validate_type == "minimum":
                                    validators.append(Number(minimum=validate["value"]))
                                elif validate_type == "maximum":
                                    validators.append(Number(maximum=validate["value"]))
                            with self.prevent(Input.Changed):
                                yield Input(
                                    str(integer_value),
                                    type="integer",
                                    classes="input",
                                    name=setting.key,
                                    validators=validators,
                                )
                        elif setting.type == "number":
                            try:
                                integer_value = float(value)
                            except (ValueError, TypeError):
                                integer_value = setting.default
                            setting_validate = setting.validate or []
                            validators: list[Validator] = []
                            for validate in setting_validate:
                                validate_type = validate["type"]
                                if validate_type == "minimum":
                                    validators.append(Number(minimum=validate["value"]))
                                elif validate_type == "maximum":
                                    validators.append(Number(maximum=validate["value"]))
                            with self.prevent(Input.Changed):
                                yield Input(
                                    str(integer_value),
                                    type="number",
                                    classes="input",
                                    name=setting.key,
                                    validators=validators,
                                )
                        elif setting.type == "choices":
                            select_value = str(value)
                            choices = setting.choices or []
                            with self.prevent(Select.Changed):
                                select_choices = [
                                    (
                                        choice
                                        if isinstance(choice, tuple)
                                        else (choice, choice)
                                    )
                                    for choice in choices
                                ]
                                choices_set = {choice[1] for choice in select_choices}
                                yield Select(
                                    select_choices,
                                    value=(
                                        select_value
                                        if select_value in choices_set
                                        else setting.default
                                    ),
                                    classes="input",
                                    name=setting.key,
                                    allow_blank=setting.default is None,
                                )

        with containers.Vertical(id="contents"):
            with containers.VerticalGroup(classes="search-container"):
                yield Input(id="search", placeholder="Search settings")
            with lazy.Reveal(
                containers.VerticalScroll(can_focus=False, id="settings-container")
            ):
                yield from compose(self, schema_to_widget("", schema.settings_map))

        yield Footer()

    @on(Input.Blurred, "Input")
    @on(Input.Submitted, "Input")
    def on_input_blurred(self, event: Input.Blurred) -> None:
        if event.validation_result and not event.validation_result.is_valid:
            self.notify(
                event.validation_result.failures[0].description or "error",
                title=event.input.name or "",
                severity="error",
            )
            event.input.value = str(
                self.app.settings.get(event.input.name or "", expand=False)
            )
            return
        if event.input.name is not None:
            if event.input.type == "integer":
                self.app.settings.set(event.input.name, int(event.value or "0"))
            elif event.input.type == "number":
                self.app.settings.set(event.input.name, float(event.value or "0"))
            else:
                self.app.settings.set(event.input.name, event.value)

    @on(TextArea.Changed)
    def on_text_area_changed(self, event: TextArea.Changed) -> None:
        if event.text_area.name is not None:
            self.app.settings.set(event.text_area.name, event.text_area.text)

    @on(Checkbox.Changed)
    def on_checkbox_changed(self, event: Checkbox.Changed) -> None:
        if event.checkbox.name is not None:
            self.app.settings.set(event.checkbox.name, event.checkbox.value)

    @on(Select.Changed)
    def on_select_changed(self, event: Select.Changed) -> None:
        if event.select.name is not None:
            self.app.settings.set(event.select.name, event.select.value)

    def filter_settings(self, search_term: str) -> None:
        if search_term:
            search_term = search_term.lower()
            for setting in self.query(".setting"):
                if setting.name:
                    setting.display = search_term in setting.name
            for container in reversed(self.query(".setting-object")):
                container.display = not container.get_child_by_id(
                    "setting-group"
                ).is_empty
        else:
            self.query(".setting").set(display=True)
            self.query(".setting-object").set(display=True)

    @on(Input.Changed, "#search")
    def on_search_input(self, event: Input.Changed) -> None:
        self.filter_settings(event.value)

    def check_action(self, action: str, parameters: tuple[object, ...]) -> bool | None:
        if action == "focus":
            if not self.is_mounted:
                return None
            return None if self.search_input.has_focus else True
        return True

    async def action_dismiss(self, result: ScreenResultType | None = None) -> None:
        self.query("#search").focus()
        self.call_after_refresh(self.dismiss, result)



import os
from importlib.metadata import version
import platform
from string import Template

from toad.app import ToadApp
from toad import paths
from toad import get_version

ABOUT_TEMPLATE = Template(
    """\
# About Toad v${TOAD_VERSION}

© Will McGugan.
                          
Toad is licensed under the terms of the [GNU AFFERO GENERAL PUBLIC LICENSE](https://www.gnu.org/licenses/agpl-3.0.txt).


## Config

config read from `$SETTINGS_PATH`
                                       
```json
$CONFIG                       
```

## Paths

| Name | Path |
| --- | --- |
| App data | `$DATA_PATH` |
| App state | `$STATE_PATH` |
| App logs | `$LOG_PATH` |

                          
## System

| System | Version |
| --- | --- |
| Python | $PYTHON |
| OS | $PLATFORM |
| Terminal | $TERMINAL |

## Dependencies

| Library | Version |
| --- | --- | 
| Toad | $TOAD_VERSION |
| Textual | $TEXTUAL_VERSION |
| Rich | $RICH_VERSION |
                          
## Environment

| Environment variable | Value |                
| --- | --- |
| `SHELL` | $SHELL |
| `TERM` | $TERM |
| `COLORTERM` | $COLORTERM |
| `TERM_PROGRAM` | $TERM_PROGRAM |
| `TERM_PROGRAM_VERSION` | $TERM_PROGRAM_VERSION |
"""
)


def render(app: ToadApp) -> str:
    """Render about markdown.

    Returns:
        Markdown string.
    """

    try:
        config: str | None = app.settings_path.read_text()
    except Exception:
        config = None

    template_data = {
        "COLORTERM": os.environ.get("COLORTERM", ""),
        "CONFIG": config,
        "DATA_PATH": paths.get_data(),
        "LOG_PATH": paths.get_log(),
        "STATE_PATH": paths.get_state(),
        "PLATFORM": platform.platform(),
        "PYTHON": f"{platform.python_implementation()} {platform.python_version()}",
        "RICH_VERSION": version("rich"),
        "SETTINGS_PATH": str(app.settings_path),
        "SHELL": os.environ.get("SHELL", ""),
        "TERM_PROGRAM_VERSION": os.environ.get("TERM_PROGRAM_VERSION", ""),
        "TERM_PROGRAM": os.environ.get("TERM_PROGRAM", ""),
        "TERM": os.environ.get("TERM", ""),
        "TEXTUAL_VERSION": version("textual"),
        "TOAD_VERSION": get_version(),
        "TERMINAL": app.term_program,
    }
    return ABOUT_TEMPLATE.safe_substitute(template_data)



from textual.app import ComposeResult
from textual.binding import Binding
from textual.events import ScreenResume
from textual.screen import ModalScreen
from textual import getters
from textual.widget import Widget
from textual import widgets
from textual import containers
from textual import on


from toad.app import ToadApp
from toad.widgets.grid_select import GridSelect
from toad.widgets.session_grid_select import SessionGridSelect
from toad.widgets.session_summary import SessionSummary


INSTRUCTIONS_NO_SESSIONS = "Your sessions will be shown here."


class SessionsScreen(ModalScreen[str]):
    CSS_PATH = "sessions.tcss"
    BINDINGS = [Binding("escape", "dismiss", "Dismiss")]

    app: getters.app[ToadApp] = getters.app(ToadApp)
    session_grid_select = getters.query_one(SessionGridSelect)

    def compose(self) -> ComposeResult:
        with containers.Center(id="title-container"):
            yield widgets.Label("Sessions")
        yield widgets.Static(INSTRUCTIONS_NO_SESSIONS, classes="instructions")
        yield SessionGridSelect(self.app.session_tracker)
        yield widgets.Footer()

    @property
    def focus_chain(self) -> list[Widget]:
        return [self.session_grid_select]

    def _on_screen_resume(self, event: ScreenResume) -> None:
        current_mode = self.app.screen_stack[0].id
        for instructions in self.query(".instructions"):
            instructions.display = not self.session_grid_select.children
        if current_mode is not None:
            self.session_grid_select.update_current(current_mode)

    def _on_screen_suspend(self) -> None:
        current_mode = self.app.screen_stack[0].id
        if current_mode is not None:
            self.session_grid_select.update_current(current_mode)

    @on(GridSelect.Selected)
    def on_selected(self, event: GridSelect.Selected) -> None:
        if (
            isinstance(event.widget, SessionSummary)
            and event.widget.session_details is not None
        ):
            self.dismiss(event.widget.session_details.mode_name)



from textual import on, work
from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual import widgets
from textual import containers
from textual import getters


class CommandEditModal(ModalScreen[str | None]):
    BINDINGS = [
        ("ctrl+c", "copy", "Copy to clipboard"),
        ("escape", "dismiss", "Dismiss"),
    ]
    AUTO_FOCUS = "Button"

    text_area = getters.query_one(widgets.TextArea)

    def __init__(
        self,
        command: str,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ):
        self.command = command
        super().__init__(name=name, id=id, classes=classes)

    def compose(self) -> ComposeResult:
        with containers.VerticalGroup(id="container"):
            with containers.Center():
                yield widgets.Label(
                    "[b]Run command",
                    classes="instructions",
                )
            yield widgets.Static(
                "Toad will run the following command (edit if you need to).",
                classes="instructions",
            )
            yield widgets.TextArea(
                self.command,
                language="bash",
                highlight_cursor_line=False,
                soft_wrap=False,
            )
            with containers.HorizontalGroup(id="button-container"):
                yield widgets.Button("OK", variant="primary", id="ok")
                yield widgets.Button(
                    "Cancel", id="cancel", action="screen.dismiss(None)"
                )
        yield widgets.Footer()

    @on(widgets.Button.Pressed, "#ok")
    def on_ok_pressed(self, event: widgets.Button.Pressed) -> None:
        self.dismiss(self.text_area.text)

    @on(widgets.Button.Pressed, "#cancel")
    def on_cancel_pressed(self, event: widgets.Button.Pressed) -> None:
        self.dismiss(None)

    def action_copy(self) -> None:
        self.app.copy_to_clipboard(self.text_area.text)
        self.notify("Command copied to clipboard", title="Copy")


if __name__ == "__main__":
    from textual.app import App

    class ModalApp(App):
        @work
        async def on_mount(self) -> None:
            result = await self.push_screen_wait(CommandEditModal("ls -al"))
            self.notify(str(result))

    ModalApp().run()



from textual.widgets import Input

from toad.directory_suggester import DirectorySuggester


class DirectoryInput(Input):
    def on_mount(self) -> None:
        self.suggester = DirectorySuggester()



from __future__ import annotations


import asyncio
import concurrent.futures

from operator import itemgetter
import os
from pathlib import Path

from typing import Sequence


from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual import work
from textual import getters
from textual import containers
from textual import events
from textual.actions import SkipAction

from textual.cache import LRUCache
from textual.reactive import var, Initialize
from textual.content import Content, Span
from textual.strip import Strip
from textual.style import Style
from textual.widget import Widget
from textual import widgets
from textual.visual import RenderOptions
from textual.widgets import OptionList, Input, DirectoryTree
from textual.widgets.option_list import Option

from toad import directory
from toad.fuzzy_index import FuzzyIndex
from toad.messages import Dismiss, InsertPath, PromptSuggestion
from toad.path_filter import PathFilter
from toad.widgets.project_directory_tree import ProjectDirectoryTree
from toad._path_fuzzy_search import PathFuzzySearch
from toad._path_match import match_path


class PathContent(Content):

    def render_strips(
        self, width: int, height: int | None, style: Style, options: RenderOptions
    ) -> list[Strip]:
        """Render the Visual into an iterable of strips. Part of the Visual protocol.

        Args:
            width: Width of desired render.
            height: Height of desired render or `None` for any height.
            style: The base style to render on top of.
            options: Additional render options.

        Returns:
            An list of Strips.
        """
        if not width:
            return []

        line = self
        if line.cell_length > width:
            while line.cell_length >= width - 3 and "/" in line.plain:
                line = line[line.plain.find("/") + 1 :]
            line = Content.assemble(("⋯ ", "$text-error"), line)

        lines = line._wrap_and_format(
            width,
            tab_size=8,
            overflow="clip",
            no_wrap=True,
            selection=options.selection,
            selection_style=options.selection_style,
            post_style=options.post_style,
            get_style=options.get_style,
        )

        if height is not None:
            lines = lines[:height]

        strip_lines = [Strip(*line.to_strip(style)) for line in lines]
        return strip_lines


class FuzzyPathOptionList(OptionList):
    """Option list with loading indicator override."""

    def get_loading_widget(self) -> Widget:
        from textual.widgets import LoadingIndicator

        return LoadingIndicator()


class FuzzyInput(Input):
    """Adds a Content placeholder to fuzzy input.

    TODO: Add this ability to Textual.
    """

    HELP = """\
## Fuzzy search

Type a few characters from the file you are searching for.

The search is *fuzzy*, and will match characters that aren't neccesarily next to each other—only the order matters.
"""

    def render_line(self, y: int) -> Strip:
        if y == 0 and not self.value:
            placeholder = Content.from_markup(self.placeholder).expand_tabs()
            placeholder = placeholder.stylize(self.visual_style)
            placeholder = placeholder.stylize(
                self.get_visual_style("input--placeholder")
            )
            if self.has_focus:
                cursor_style = self.get_visual_style("input--cursor")
                if self._cursor_visible:
                    # If the placeholder is empty, there's no characters to stylise
                    # to make the cursor flash, so use a single space character
                    if len(placeholder) == 0:
                        placeholder = Content(" ")
                    placeholder = placeholder.stylize(cursor_style, 0, 1)

            strip = Strip(placeholder.render_segments())
            return strip

        return super().render_line(y)


class PathSearch(containers.VerticalGroup):

    BINDING_GROUP_TITLE = "Path search"

    CURSOR_BINDING_GROUP = Binding.Group(description="Move selection")
    BINDINGS = [
        Binding(
            "up", "cursor_up", "Cursor up", group=CURSOR_BINDING_GROUP, priority=True
        ),
        Binding(
            "down",
            "cursor_down",
            "Cursor down",
            group=CURSOR_BINDING_GROUP,
            priority=True,
        ),
        Binding("enter", "submit", "Insert path", priority=True, show=False),
        Binding("escape", "dismiss", "Dismiss", priority=True, show=False),
        Binding("tab", "switch_picker", "Switch picker", priority=True, show=False),
    ]

    def get_fuzzy_search(self) -> PathFuzzySearch:
        return PathFuzzySearch(case_sensitive=False)

    root: var[Path] = var(Path("./"))
    paths: var[list[Path]] = var(list)
    display_paths: var[list[str]] = var(list)
    filtered_path_indices: var[list[int]] = var(list)
    loaded = var(False)
    filter = var("")
    fuzzy_search: var[PathFuzzySearch] = var(Initialize(get_fuzzy_search))
    show_tree_picker: var[bool] = var(False)

    option_list = getters.query_one(FuzzyPathOptionList)
    tree_view = getters.query_one(ProjectDirectoryTree)
    input = getters.query_one(Input)

    def __init__(self, root: Path) -> None:
        super().__init__()
        self.set_reactive(PathSearch.root, root)
        self.root = root
        self.fuzzy_index = FuzzyIndex()
        self.pool = concurrent.futures.InterpreterPoolExecutor(
            thread_name_prefix=f"fuzzy-path-search-{root}"
        )
        self.search_cache: LRUCache[str, list[tuple[float, Sequence[int], str]]] = (
            LRUCache(1024)
        )

    def compose(self) -> ComposeResult:
        with widgets.ContentSwitcher(initial="path-search-fuzzy"):
            with containers.VerticalGroup(id="path-search-fuzzy"):
                yield FuzzyInput(
                    compact=True, placeholder="fuzzy search \t[r]▌tab▐[/r] tree view"
                )
                yield FuzzyPathOptionList()
            with containers.VerticalGroup(id="path-search-tree"):
                yield widgets.Static(
                    Content.from_markup(
                        "tree view \t[r]▌tab▐[/] fuzzy search"
                    ).expand_tabs(),
                    classes="message",
                )
                yield ProjectDirectoryTree(self.root).data_bind(path=PathSearch.root)

    def on_mount(self) -> None:
        tree = self.tree_view
        tree.guide_depth = 2
        tree.center_scroll = True

    def watch_show_tree_picker(self, show_tree_picker: bool) -> None:
        content_switcher = self.query_one(widgets.ContentSwitcher)
        content_switcher.current = (
            "path-search-tree" if show_tree_picker else "path-search-fuzzy"
        )
        if show_tree_picker:
            self.tree_view.focus()

        else:
            self.input.focus()

    def action_switch_picker(self) -> None:
        self.show_tree_picker = not self.show_tree_picker

    def fuzzy_match_paths(
        self, search: str, paths: list[str]
    ) -> list[tuple[float, Sequence[int], str]]:

        scores = list(
            self.pool.map(
                match_path,
                [(search, path) for path in paths],
                chunksize=10,
            )
        )
        return scores

    async def search(self, search: str) -> None:
        if not search:
            self.option_list.set_options(
                [
                    Option(self.highlight_path(path), path)
                    for path in self.display_paths[:100]
                ],
            )
            return

        display_paths = await self.fuzzy_index.search(search)

        if len(display_paths) > 20:
            if (scored_paths := self.search_cache.get(search)) is None:
                scored_paths = await asyncio.to_thread(
                    self.fuzzy_match_paths, search, display_paths
                )
                self.search_cache[search] = scored_paths
        else:
            fuzzy_search = self.fuzzy_search
            scored_paths: list[tuple[float, Sequence[int], str]] = [
                (
                    *fuzzy_search.match(search, path),
                    path,
                )
                for path in display_paths
            ]

        scored_paths = sorted(
            [score for score in scored_paths if score[0]],
            key=itemgetter(0),
            reverse=True,
        )

        scores = [
            (score, highlights, self.highlight_path(path))
            for score, highlights, path in scored_paths[:30]
        ]

        def highlight_offsets(path: Content, offsets: Sequence[int]) -> Content:
            highlighted_path = path.add_spans(
                [Span(offset, offset + 1, "underline") for offset in offsets]
            )
            return PathContent(
                highlighted_path.plain,
                list(highlighted_path.spans),
                highlighted_path.cell_length,
            )

        self.option_list.set_options(
            [
                Option(highlight_offsets(path, offsets), id=path.plain)
                for index, (score, offsets, path) in enumerate(scores)
            ]
        )
        with self.option_list.prevent(OptionList.OptionHighlighted):
            self.option_list.highlighted = 0
        self.post_message(PromptSuggestion(""))

    def action_cursor_down(self) -> None:
        if self.show_tree_picker:
            self.tree_view.action_cursor_down()
        else:
            self.option_list.action_cursor_down()

    def action_cursor_up(self) -> None:
        if self.show_tree_picker:
            self.tree_view.action_cursor_up()
        else:
            self.option_list.action_cursor_up()

    def action_dismiss(self) -> None:
        self.post_message(Dismiss(self))
        self.filter = ""

    def on_show(self) -> None:
        self.focus()

    def focus(self, scroll_visible: bool = False) -> Self:
        if self.show_tree_picker:
            return self.tree_view.focus(scroll_visible=scroll_visible)
        else:
            return self.input.focus(scroll_visible=scroll_visible)

    def on_descendant_blur(self, event: events.DescendantBlur) -> None:
        if self.show_tree_picker:
            if event.widget == self.tree_view:
                self.post_message(Dismiss(self))
        else:
            if event.widget == self.input:
                self.post_message(Dismiss(self))

    @classmethod
    def make_relative(cls, path: Path, root: Path) -> Path:
        """Make a path relative from the root.

        Args:
            path: Path to consider.
            root: Root path.

        Returns:
            A relative path.
        """
        return path.resolve().relative_to(root.resolve())

    @on(DirectoryTree.NodeHighlighted)
    async def on_node_highlighted(self, event: DirectoryTree.NodeHighlighted) -> None:
        event.stop()

        dir_entry = event.node.data
        if dir_entry is not None:
            try:
                path = await asyncio.to_thread(
                    self.make_relative, dir_entry.path, self.root
                )
            except ValueError:
                # Being defensive here, shouldn't occur
                return
            tree_path = str(path)
            self.post_message(PromptSuggestion(tree_path))

    @on(DirectoryTree.FileSelected)
    async def on_file_selected(self, event: DirectoryTree.FileSelected) -> None:
        event.stop()

        dir_entry = event.node.data
        if dir_entry is not None:
            try:
                path = await asyncio.to_thread(
                    self.make_relative, dir_entry.path, self.root
                )
            except ValueError:
                return
            tree_path = str(path)
            self.post_message(InsertPath(tree_path))
            self.post_message(Dismiss(self))

    @on(Input.Changed)
    async def on_input_changed(self, event: Input.Changed):
        await self.search(event.value)

    @on(OptionList.OptionHighlighted)
    async def on_option_list_changed(self, event: OptionList.OptionHighlighted):
        event.stop()
        if event.option:
            self.post_message(PromptSuggestion(event.option.id))

    @on(OptionList.OptionSelected)
    def on_option_list_option_selected(self, event: OptionList.OptionSelected) -> None:
        self.action_submit()

    def action_submit(self):
        if self.show_tree_picker:
            raise SkipAction()

        elif (highlighted := self.option_list.highlighted) is not None:
            option = self.option_list.options[highlighted]
            if option.id:
                self.post_message(InsertPath(option.id))
                self.post_message(Dismiss(self))

    def get_path_filter(self, project_path: Path) -> PathFilter:
        """Get a PathFilter insance for the give project path.

        Args:
            project_path: Project path.

        Returns:
            `PathFilter` object.
        """
        path_filter = PathFilter.from_git_root(project_path)
        return path_filter

    def reset(self) -> None:
        """Reset and focus input."""
        self.input.clear()
        self.input.focus()

    @work(exclusive=True)
    async def refresh_paths(self):
        self.option_list.set_loading(True)
        root = self.root
        try:
            path_filter = await asyncio.to_thread(self.get_path_filter, root)
            self.tree_view.path_filter = path_filter
            self.tree_view.clear()
            self.tree_view.reload()
            paths = await directory.scan(
                root, path_filter=path_filter, add_directories=True
            )

            def make_absolute(paths: list[Path]) -> list[Path]:
                """Make all paths absolute.

                Args:
                    paths: A list of paths.

                Returns:
                    List of absolute paths,

                """
                return [path.absolute() for path in paths]

            paths = await asyncio.to_thread(make_absolute, paths)
            self.root = root
            self.paths = paths
        except Exception:
            self.option_list.set_loading(False)
            raise

    def highlight_path(self, path: str) -> PathContent:
        content = Content.styled(path, "$text 50%")
        if os.path.split(path)[-1].startswith("."):
            return PathContent(
                content.plain, list(content.spans), cell_length=content.cell_length
            )
        content = content.highlight_regex("[^/]*?$", style="$text-primary")
        content = content.highlight_regex(r"\.[^/]*$", style="italic")
        return PathContent(
            content.plain, list(content.spans), cell_length=content.cell_length
        )

    @work(description="watch_paths")
    async def watch_paths(self, paths: list[Path]) -> None:

        def path_display(path: Path) -> str:
            try:
                is_directory = path.is_dir()
            except OSError:
                is_directory = False
            if is_directory:
                return str(path.relative_to(self.root)) + "/"
            else:
                return str(path.relative_to(self.root))

        def make_display_paths() -> list[str]:
            display_paths = sorted(map(path_display, paths), key=str.lower)
            display_paths.sort(key=lambda path: path.count("/"))
            return display_paths

        self.display_paths = await asyncio.to_thread(make_display_paths)

        self.option_list.highlighted = None
        self._update_paths(self.display_paths)

        self.option_list.set_options(
            [
                Option(self.highlight_path(path), id=path)
                for path in self.display_paths[:100]
            ]
        )
        with self.option_list.prevent(OptionList.OptionHighlighted):
            self.option_list.highlighted = 0

        self.post_message(PromptSuggestion(""))

    @work(description="update_paths")
    async def _update_paths(self, paths: list[str]) -> None:
        """Update the paths index.

        Args:
            paths: A list of paths.
        """
        await self.fuzzy_index.update_paths(paths)
        self.call_after_refresh(self.option_list.set_loading, False)



from textual.app import ComposeResult
from textual import getters
from textual import containers
from textual import widgets
from textual.reactive import reactive
from textual.timer import Timer


from toad.widgets.condensed_path import CondensedPath
from toad.widgets.throbber import ThrobberVisual
from toad.session_tracker import SessionDetails


class BusyIndicator(widgets.Static):
    def render(self) -> ThrobberVisual:
        return ThrobberVisual("▔")

    def on_mount(self) -> None:
        self.auto_refresh = 1 / 4


class SessionSummary(containers.VerticalGroup):
    session_details: reactive[SessionDetails | None] = reactive(
        None, always_update=True, recompose=True
    )

    title = getters.query_one(".title", widgets.Label)
    subtitle = getters.query_one(".subtitle", widgets.Label)
    current = reactive(True, toggle_class="-current", init=False, recompose=True)
    blink = reactive(False, toggle_class="-blink")

    def __init__(
        self,
        session_details: SessionDetails | None,
        id: str | None = None,
        classes: str | None = None,
    ) -> None:
        super().__init__(id=id, classes=classes)
        self.blink_timer: Timer | None = None
        self.set_reactive(SessionSummary.session_details, session_details)

    def on_mount(self) -> None:

        def do_blink() -> None:
            self.blink = not self.blink

        self.blink_timer = self.set_interval(0.5, do_blink, pause=False)

    def watch_current(self, current: bool) -> None:
        if (session_details := self.session_details) is not None:
            if current:
                self.title.update(f"✱{session_details.title}")
            else:
                self.title.update(session_details.title)

    def compose(self) -> ComposeResult:
        if (session_details := self.session_details) is not None:
            self.remove_class(
                "-state-notready",
                "-state-busy",
                "-state-asking",
                "-state-idle",
                update=False,
            )
            self.add_class(f"-state-{session_details.state}")

        yield BusyIndicator()
        # yield widgets.Rule(line_style="heavy")
        with containers.HorizontalGroup():
            yield widgets.Label("❯", classes="icon")
            with containers.VerticalGroup():
                if session_details is not None:
                    yield widgets.Label(
                        (
                            f"✱{session_details.title}"
                            if self.current
                            else session_details.title
                        ),
                        classes="title",
                        markup=False,
                    )
                    yield widgets.Label(
                        session_details.subtitle,
                        classes="subtitle",
                        markup=False,
                    )
                    with containers.HorizontalGroup():
                        yield widgets.Label("📁 ")
                        yield CondensedPath(session_details.path)


if __name__ == "__main__":
    from textual.app import App

    session_details = SessionDetails(
        0,
        "mode",
        title="Building BMI calculator",
        subtitle="Claude Code",
    )

    class SessionApp(App):
        CSS_PATH = "../toad.tcss"
        CSS = """
        Screen {
            layout: horizontal;
        }
        """

        def compose(self) -> ComposeResult:
            yield SessionSummary(session_details, classes="-state-busy")
            yield SessionSummary(session_details, classes="-state-asking")
            yield SessionSummary(session_details, classes="-state-idle")

    SessionApp().run()



from typing import cast

from textual import on
from textual import getters
from textual.app import ComposeResult

from textual import work
from textual.content import Content
from textual.screen import ModalScreen
from textual import containers
from textual import widgets
from textual.reactive import var

import toad
from textual.binding import Binding
from toad.agent_schema import Action, Agent, OS, Command
from toad.app import ToadApp


class DescriptionContainer(containers.VerticalScroll):
    def allow_focus(self) -> bool:
        """Focus only if it can be scrolled."""
        return self.show_vertical_scrollbar


class AgentModal(ModalScreen):
    AUTO_FOCUS = "Select#action-select"

    BINDINGS = [
        Binding("escape", "dismiss(None)", "Dismiss", show=False),
        Binding("space", "dismiss('launch')", "Launch agent", priority=True),
    ]

    action = var("")

    app = getters.app(ToadApp)
    action_select = getters.query_one("#action-select", widgets.Select)
    launcher_checkbox = getters.query_one("#launcher-checkbox", widgets.Checkbox)

    def __init__(self, agent: Agent) -> None:
        self._agent = agent
        super().__init__()

    def compose(self) -> ComposeResult:
        launcher_set = frozenset(
            self.app.settings.get("launcher.agents", str).splitlines()
        )

        agent = self._agent

        app = self.app
        launcher_set = frozenset(app.settings.get("launcher.agents", str).splitlines())
        agent = self._agent
        actions = agent["actions"]

        script_os = cast(OS, toad.os)
        if script_os not in actions:
            script_os = "*"

        commands: dict[Action, Command] = actions[cast(OS, script_os)]
        script_choices = [
            (action["description"], name) for name, action in commands.items()
        ]
        script_choices.append((f"Launch {agent['name']}", "__launch__"))

        with containers.Vertical(id="container"):
            with DescriptionContainer(id="description-container"):
                yield widgets.Markdown(agent["help"], id="description")
            with containers.VerticalGroup():
                if "install_acp" in commands:
                    yield widgets.Static(
                        Content(
                            f"{agent['name']} requires an ACP adapter to work with Toad. Install from the actions list."
                        ),
                        classes="acp-warning",
                    )
                with containers.HorizontalGroup():
                    yield widgets.Checkbox(
                        "Show in launcher",
                        value=agent["identity"] in launcher_set,
                        id="launcher-checkbox",
                    )
                    yield widgets.Select(
                        script_choices,
                        prompt="Actions",
                        allow_blank=True,
                        id="action-select",
                    )
                    yield widgets.Button(
                        "Go", variant="primary", id="run-action", disabled=True
                    )
        yield widgets.Footer()

    def on_mount(self) -> None:
        self.query_one("Footer").styles.animate("opacity", 1.0, duration=500 / 1000)

    @on(widgets.Checkbox.Changed)
    def on_checkbox_changed(self, event: widgets.Select.Changed) -> None:
        launcher_agents = self.app.settings.get("launcher.agents", str).splitlines()
        agent_identity = self._agent["identity"]
        if agent_identity in launcher_agents:
            launcher_agents.remove(agent_identity)
        if event.value:
            launcher_agents.insert(0, agent_identity)
        self.app.settings.set("launcher.agents", "\n".join(launcher_agents))

    @on(widgets.Select.Changed)
    def on_select_changed(self, event: widgets.Select.Changed) -> None:
        self.action = event.value if isinstance(event.value, str) else ""

    @work
    @on(widgets.Button.Pressed)
    async def on_button_pressed(self) -> None:
        agent = self._agent
        action = self.action_select.value

        assert isinstance(action, str)
        if action == "__launch__":
            self.dismiss("launch")
            return

        agent_actions = self._agent["actions"]

        if (commands := agent_actions.get(toad.os, None)) is None:
            commands = agent_actions.get("*", None)
        if commands is None:
            self.notify(
                "Action is not available on this platform",
                title="Agent action",
                severity="error",
            )
            return
        command = commands[action]

        from toad.screens.action_modal import ActionModal
        from toad.screens.command_edit_modal import CommandEditModal

        title = command["description"]
        agent_id = self._agent["identity"]
        action_command = command["command"]
        bootstrap_uv = command.get("bootstrap_uv", False)

        agent = self._agent
        # Focus the select
        # It's unlikely the user wants to re-run the action
        self.action_select.focus()

        action_command = await self.app.push_screen_wait(
            CommandEditModal(action_command)
        )
        if action_command is None:
            return

        return_code = await self.app.push_screen_wait(
            ActionModal(
                action,
                agent_id,
                title,
                action_command,
                bootstrap_uv=bootstrap_uv,
            )
        )
        if return_code == 0 and action in {"install", "install-acp"}:
            # Add to launcher if we installed something
            if not self.launcher_checkbox.value:
                self.notify(
                    f"{agent['name']} has been added to your launcher",
                    title="Add agent",
                    severity="information",
                )
                self.launcher_checkbox.value = True

    def watch_action(self, action: str) -> None:
        go_button = self.query_one("#run-action", widgets.Button)
        go_button.disabled = not action
